"""
@file
@brief Metrics to compare machine learning.
"""
import numpy
from sklearn.metrics import r2_score

_known_functions = {
    'exp': numpy.exp,
    'log': numpy.log
}


def comparable_metric(metric_function, y_true, y_pred,
                      tr="log", inv_tr='exp', **kwargs):
    """
    Applies function on either the true target or/and the predictions
    before computing r2 score.

    :param metric_function: metric to compute
    :param y_true: expected targets
    :param y_pred: predictions
    :param sample_weight: weights
    :param multioutput: see :epkg:`sklearn:metrics:r2_score`
    :param tr: transformation applied on the target
    :param inv_tr: transformation applied on the predictions
    :return: results
    """
    tr = _known_functions.get(tr, tr)
    inv_tr = _known_functions.get(inv_tr, inv_tr)
    if tr is not None and not callable(tr):
        raise TypeError("Argument tr must be callable.")
    if inv_tr is not None and not callable(inv_tr):
        raise TypeError("Argument inv_tr must be callable.")
    if tr is None and inv_tr is None:
        raise ValueError(
            "tr and inv_tr cannot be both None at the same time.")
    if tr is None:
        return metric_function(y_true, inv_tr(y_pred), **kwargs)
    if inv_tr is None:
        return metric_function(tr(y_true), y_pred, **kwargs)
    return metric_function(tr(y_true), inv_tr(y_pred), **kwargs)


def r2_score_comparable(y_true, y_pred, *, sample_weight=None,
                        multioutput='uniform_average',
                        tr=None, inv_tr=None):
    """
    Applies function on either the true target or/and the predictions
    before computing r2 score.

    :param y_true: expected targets
    :param y_pred: predictions
    :param sample_weight: weights
    :param multioutput: see :epkg:`sklearn:metrics:r2_score`
    :param tr: transformation applied on the target
    :param inv_tr: transformation applied on the predictions
    :return: results

    Example:

    .. runpython::
        :showcode:

        import numpy
        from sklearn import datasets
        from sklearn.model_selection import train_test_split
        from sklearn.linear_model import LinearRegression
        from sklearn.metrics import r2_score
        from mlinsights.metrics import r2_score_comparable

        iris = datasets.load_iris()
        X = iris.data[:, :4]
        y = iris.target + 1

        X_train, X_test, y_train, y_test = train_test_split(X, y)

        model1 = LinearRegression().fit(X_train, y_train)
        print('r2', r2_score(y_test, model1.predict(X_test)))
        print('r2 log', r2_score(numpy.log(y_test), numpy.log(model1.predict(X_test))))
        print('r2 log comparable', r2_score_comparable(
            y_test, model1.predict(X_test), tr="log", inv_tr="log"))

        model2 = LinearRegression().fit(X_train, numpy.log(y_train))
        print('r2', r2_score(numpy.log(y_test), model2.predict(X_test)))
        print('r2 log', r2_score(y_test, numpy.exp(model2.predict(X_test))))
        print('r2 log comparable', r2_score_comparable(
            y_test, model2.predict(X_test), inv_tr="exp"))
    """
    return comparable_metric(r2_score, y_true, y_pred,
                             sample_weight=sample_weight,
                             multioutput=multioutput,
                             tr=tr, inv_tr=inv_tr)



"""
@file
@brief Access to the C API of scikit-learn (decision tree)
"""
from libc.stdio cimport printf

import numpy
cimport numpy
numpy.import_array()

ctypedef numpy.npy_intp SIZE_t

from sklearn.tree._tree cimport Tree

TREE_LEAF = -1
TREE_UNDEFINED = -2


cdef SIZE_t _tree_add_node(Tree tree,
                           SIZE_t parent,
                           bint is_left,
                           bint is_leaf,
                           SIZE_t feature,
                           double threshold,
                           double impurity,
                           SIZE_t n_node_samples,
                           double weighted_n_node_samples):
    if parent == -1:
        parent = TREE_UNDEFINED
    return tree._add_node(parent, is_left, is_leaf, feature,
                          threshold, impurity,
                          n_node_samples, weighted_n_node_samples)


def tree_add_node(tree, parent, is_left, is_leaf, feature, threshold,
                  impurity, n_node_samples, weighted_n_node_samples):
    """
    Adds a node to tree.

    :param parent: parent index (-1 for the root)
    :param is_left: is left node?
    :param is_leaf: is leave?
    :param feature: feature index
    :param threshold: threshold (or value)
    :param impurity: impurity
    :param n_node_samples: number of samples this node represents
    :param weighted_n_node_samples: node weight
    """
    return _tree_add_node(tree, parent, is_left, is_leaf, feature, threshold,
                          impurity, n_node_samples, weighted_n_node_samples)



"""
@file
@brief Helpers to investigate a tree structure.

.. versionadded:: 0.4
"""
import numpy
from sklearn.tree._tree import Tree  # pylint: disable=E0611
from sklearn.tree import DecisionTreeRegressor
from ._tree_digitize import tree_add_node  # pylint: disable=E0611


def digitize2tree(bins, right=False):
    """
    Builds a decision tree which returns the same result as
    `lambda x: numpy.digitize(x, bins, right=right)`
    (see :epkg:`numpy:digitize`).

    :param bins: array of bins. It has to be 1-dimensional and monotonic.
    :param right: Indicating whether the intervals include the right
        or the left bin edge. Default behavior is (right==False)
        indicating that the interval does not include the right edge.
        The left bin end is open in this case, i.e.,
        `bins[i-1] <= x < bins[i]` is the default behavior for
        monotonically increasing bins.
    :return: decision tree

    .. note::
        The implementation of decision trees in :epkg:`scikit-learn`
        only allows one type of decision (`<=`). That's why the
        function throws an exception when `right=False`. However,
        this could be overcome by using :epkg:`ONNX` where all
        kind of decision rules are implemented. Default value for
        right is still *False* to follow *numpy* API even though
        this value raises an exception in *digitize2tree*.

    The following example shows what the tree looks like.

    .. runpython::
        :showcode:

        import numpy
        from sklearn.tree import export_text
        from mlinsights.mltree import digitize2tree

        x = numpy.array([0.2, 6.4, 3.0, 1.6])
        bins = numpy.array([0.0, 1.0, 2.5, 4.0, 7.0])
        expected = numpy.digitize(x, bins, right=True)
        tree = digitize2tree(bins, right=True)
        pred = tree.predict(x.reshape((-1, 1)))
        print("Comparison with numpy:")
        print(expected, pred)
        print("Tree:")
        print(export_text(tree, feature_names=['x']))

    See also example :ref:`l-example-digitize`.

    .. versionadded:: 0.4
    """
    if not right:
        raise RuntimeError(
            "right must be True not right=%r" % right)
    ascending = len(bins) <= 1 or bins[0] < bins[1]

    if not ascending:
        bins2 = bins[::-1]
        cl = digitize2tree(bins2, right=right)
        n = len(bins)
        for i in range(cl.tree_.value.shape[0]):
            cl.tree_.value[i, 0, 0] = n - cl.tree_.value[i, 0, 0]
        return cl

    tree = Tree(1, numpy.array([1], dtype=numpy.intp), 1)
    values = []
    UNUSED = numpy.nan
    n_nodes = []

    def add_root(index):
        if index < 0 or index >= len(bins):
            raise IndexError(  # pragma: no cover
                "Unexpected index %d / len(bins)=%d." % (
                    index, len(bins)))
        parent = -1
        is_left = False
        is_leaf = False
        threshold = bins[index]
        n = tree_add_node(
            tree, parent, is_left, is_leaf, 0, threshold, 0, 1, 1.)
        values.append(UNUSED)
        n_nodes.append(n)
        return n

    def add_nodes(parent, i, j, is_left):
        # add for bins[i:j] (j excluded)
        if is_left:
            # it means j is the parent split
            if i == j:
                # leaf
                n = tree_add_node(tree, parent, is_left, True, 0, 0, 0, 1, 1.)
                n_nodes.append(n)
                values.append(i)
                return n
            if i + 1 == j:
                # split
                values.append(UNUSED)
                th = bins[i]
                n = tree_add_node(tree, parent, is_left,
                                  False, 0, th, 0, 1, 1.)
                n_nodes.append(n)
                add_nodes(n, i, i, True)
                add_nodes(n, i, j, False)
                return n
            if i + 1 < j:
                # split
                values.append(UNUSED)
                index = (i + j) // 2
                th = bins[index]
                n = tree_add_node(tree, parent, is_left,
                                  False, 0, th, 0, 1, 1.)
                n_nodes.append(n)
                add_nodes(n, i, index, True)
                add_nodes(n, index, j, False)
                return n
        else:
            # it means i is the parent split
            if i + 1 == j:
                # leaf
                values.append(j)
                n = tree_add_node(tree, parent, is_left, True, 0, 0, 0, 1, 1.)
                n_nodes.append(n)
                return n
            if i + 1 < j:
                # split
                values.append(UNUSED)
                index = (i + j) // 2
                th = bins[index]
                n = tree_add_node(tree, parent, is_left,
                                  False, 0, th, 0, 1, 1.)
                n_nodes.append(n)
                add_nodes(n, i, index, True)
                add_nodes(n, index, j, False)
                return n
        raise NotImplementedError(  # pragma: no cover
            "Unexpected case where i=%r, j=%r, is_left=%r." % (
                i, j, is_left))

    index = len(bins) // 2
    add_root(index)
    add_nodes(0, 0, index, True)
    add_nodes(0, index, len(bins), False)

    cl = DecisionTreeRegressor()
    cl.tree_ = tree
    cl.tree_.value[:, 0, 0] = numpy.array(  # pylint: disable=E1137
        values, dtype=numpy.float64)
    cl.n_outputs = 1
    cl.n_outputs_ = 1
    try:
        # scikit-learn >= 0.24
        cl.n_features_in_ = 1
    except AttributeError:
        # scikit-learn < 0.24
        cl.n_features_ = 1
    try:
        # for scikit-learn<=0.23.2
        cl.n_features_ = 1
    except AttributeError:
        pass
    return cl



# -*- coding: utf-8 -*-
import sys
import os
import warnings
from setuptools import setup, Extension, find_packages
from pyquicksetup import read_version, read_readme, default_cmdclass

#########
# settings
#########

project_var_name = "mlinsights"
versionPython = "%s.%s" % (sys.version_info.major, sys.version_info.minor)
path = "Lib/site-packages/" + project_var_name
readme = 'README.rst'
history = "HISTORY.rst"
requirements = None

KEYWORDS = [project_var_name, 'Xavier Dupré', 'machine learning',
            'scikit-learn']
DESCRIPTION = """Extends scikit-learn with a couple of new models, transformers, metrics, plotting."""
CLASSIFIERS = [
    'Programming Language :: Python :: 3',
    'Intended Audience :: Developers',
    'Topic :: Scientific/Engineering',
    'Topic :: Education',
    'License :: OSI Approved :: MIT License',
    'Development Status :: 5 - Production/Stable'
]


#######
# data
#######

packages = find_packages()
package_dir = {k: os.path.join('.', k.replace(".", "/")) for k in packages}
package_data = {
    project_var_name + ".mlmodel": ["*.pxd", "*.pyx"],
}


def get_extensions():
    root = os.path.abspath(os.path.dirname(__file__))
    if sys.platform.startswith("win"):
        extra_compile_args = None
    else:
        extra_compile_args = ['-std=c++11']

    ext_modules = []

    # mlmodel

    import sklearn
    extensions = ["direct_blas_lapack"]
    spl = sklearn.__version__.split('.')
    vskl = (int(spl[0]), int(spl[1]))
    if vskl >= (0, 24):
        extensions.append(("_piecewise_tree_regression_common",
                           "_piecewise_tree_regression_common024"))
    else:
        raise ImportError("Cannot build mlisinghts for scikit-learn<1.0.")

    extensions.extend([
        "piecewise_tree_regression_criterion",
        "piecewise_tree_regression_criterion_linear",
        "piecewise_tree_regression_criterion_fast",
        "_tree_digitize",
    ])

    pattern1 = "mlinsights.%s.%s"
    import numpy
    for name in extensions:
        folder = "mltree" if name == "_tree_digitize" else "mlmodel"
        if isinstance(name, tuple):
            m = Extension(pattern1 % (folder, name[0]),
                          ['mlinsights/%s/%s.pyx' % (folder, name[1])],
                          include_dirs=[numpy.get_include()],
                          extra_compile_args=["-O3"],
                          language='c')
        else:
            m = Extension(pattern1 % (folder, name),
                          ['mlinsights/%s/%s.pyx' % (folder, name)],
                          include_dirs=[numpy.get_include()],
                          extra_compile_args=["-O3"],
                          language='c')
        ext_modules.append(m)

    # cythonize
    from Cython.Build import cythonize
    opts = dict(boundscheck=False, cdivision=True,
                wraparound=False, language_level=3,
                cdivision_warnings=False, embedsignature=True,
                initializedcheck=False)
    ext_modules = cythonize(ext_modules, compiler_directives=opts)
    return ext_modules


try:
    ext_modules = get_extensions()
except ImportError as e:
    warnings.warn(
        "Unable to build C++ extension with missing dependencies %r." % e)
    ext_modules = None

# setup

setup(
    name=project_var_name,
    version=read_version(__file__, project_var_name),
    author='Xavier Dupré',
    author_email='xavier.dupre@gmail.com',
    license="MIT",
    url="http://www.xavierdupre.fr/app/%s/helpsphinx/index.html" % project_var_name,
    download_url="https://github.com/sdpython/%s/" % project_var_name,
    description=DESCRIPTION,
    long_description=read_readme(__file__),
    cmdclass=default_cmdclass(),
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    packages=packages,
    package_dir=package_dir,
    package_data=package_data,
    setup_requires=["pyquicksetup", 'cython', 'scipy', 'scikit-learn'],
    install_requires=['cython', 'scikit-learn>=1.0', 'pandas', 'scipy',
                      'matplotlib', 'pandas_streaming', 'numpy>=1.16'],
    ext_modules=ext_modules,  # cythonize(ext_modules),
)
